/* 
   ffmpegfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "avformat.h"
#include "nvrec.h"
#ifdef HAVE_CONVERT
#include "convert.h"
#endif

struct ffmpegfile_core_t {
	AVFormatContext * oc;
	AVFrame *pic;
	int video_stream;
	int audio_stream;
	AVCodecContext *video_cc;
	AVCodecContext *audio_cc;
	unsigned char video_buffer[128*1024];
	unsigned char * audio_buffer;
	unsigned char * audio_out;
	int abuf_used;
	int frame_bytes;
	unsigned int split_size;
	char *basename;
	char extension[4];
#ifdef HAVE_CONVERT
	convert_t * conv;
#endif
};

void ffmpegfile_core_usage()
{
	MPRINT("FFMPEG file (mpeg-1) core usage:\n");
	MPRINT("\t-t split output size (in Mbytes)         [35]\n");
	MPRINT("\t-o <output filename>                     [NONE]\n");
	MPRINT("\t * filename starting with split://base indicates base name\n");
	MPRINT("\t-ext extension to use with basename (only with split)\n");
	MPRINT("\t-ff <file format \"help\" for list>        [from ext]\n");
	MPRINT("\t-vc <video codec \"help\" for list>        [from ext]\n");
	MPRINT("\t-vb <kbits per second for video>           [1800]\n");
	MPRINT("\t-vg <video gop size - 0 for I frame only>[25]\n");
	MPRINT("\t-ac <audio codec \"help\" for list>        [from ext]\n");
	MPRINT("\t-ab <kbits per second for audio>         [80]\n");
	MPRINT("\t-qmin <parameter> [value] set video min\n");
	MPRINT("\t   quantizer <1..31>                     [2]\n");
	MPRINT("\t-qmax <parameter> [value] set video max\n");
	MPRINT("\t   quantizer <1..31>                     [10]\n");
#ifdef HAVE_CONVERT
	MPRINT("\t-vhook \"module name and params\"          [NONE]\n");
#endif
	MPRINT("\t-o <output filename - may not contain :> [NONE]\n");

}

struct file_core_t ffmpegfile_core;

void ffmpegfile_core_novideo(struct file_core_t * fct, unsigned char **buf)
{
}

void ffmpegfile_core_nonullvideo(struct file_core_t * fct)
{
}

void ffmpegfile_core_noaudio(struct file_core_t * fct, unsigned char *buf, int n)
{
}

int ffmpeg_setparm(char **args, int vid, AVCodecContext *cc)
{
 	NV_ARGS;
	
	if(vid) {
		cc->bit_rate = 1800 * 1000;
		cc->bit_rate_tolerance = 1024 * 8;
	} else {
		cc->bit_rate = 80 * 1000;
		cc->bit_rate_tolerance = 1024 * 8;
	}
	cc->flags = 0;
	cc->sub_id = 0;
	cc->me_method = 0;
	cc->gop_size = 250;
	
	cc->qcompress = 0.5;
	cc->qblur = 0.5;
	cc->qmin = 3;
	cc->qmax = 15;
	cc->max_qdiff = 3;
	cc->max_b_frames = 0;
	cc->b_quant_factor = 2;
	cc->rc_strategy = 2;		/* Obsolete - remove ? */
	cc->b_frame_strategy = 0;	/* Obsolete - remove ? */
	
	cc->rtp_mode = 0;
	cc->rtp_payload_size = 4096;
	cc->rtp_callback = NULL;
	
	cc->workaround_bugs = 1;
	cc->strict_std_compliance = -1;
	cc->error_resilience = 1;
	if(vid) {
		cc->luma_elim_threshold = 0;
		cc->chroma_elim_threshold = 0;
	}
	cc->b_quant_offset = 0;
	
	NV_SCAN_ARGS{
#define SETINT(x, min, max) if(!NV_NEXT_ARG) return -1; cc->x = atoi(NV_THIS_ARG); cc->x=(cc->x>min)?cc->x:min; cc->x=(cc->x>max)?max:cc->x
#define VIDARG(x) if((strcmp(NV_THIS_ARG, #x) == 0) && (vid))
#define AUDARG(x) if((strcmp(NV_THIS_ARG, #x) == 0) && (!vid))
#define ARG(x) if(strcmp(NV_THIS_ARG, #x) == 0)
		VIDARG(-vb) {
			SETINT(bit_rate, 4, 24000);
			cc->bit_rate *= 1000;
		} else AUDARG(-ab) {
			SETINT(bit_rate, 16, 256);
			cc->bit_rate *= 1000;
		} else ARG(-vg) {
			SETINT(gop_size, 0, 1000000);
		} else VIDARG(-qmin) {
			SETINT(qmin, 1, 31);
		} else VIDARG(-qmax) {
			SETINT(qmax, 1, 31);
		}
	}
	return 0;
}

void ffmpegfile_core_rotate(struct file_core_t * fct)
{
	struct ffmpegfile_core_t *ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	char middle[17];
	char *name_buffer;
	struct tm *ptr;
	time_t tm;
	int len;
	
	if(!ffmpegc)
		return;
	av_write_trailer(ffmpegc->oc);
	if (url_fclose(&ffmpegc->oc->pb) < 0) {
		EPRINT("Could not close filename");
		goto error0;
	}
	
	tm = time(NULL);
	ptr = gmtime(&tm);
	strftime(middle, 17, "%Y%m%dT%H%M%S", ptr);
	len = strlen(ffmpegc->basename) + 1;
	
	name_buffer = (char *)malloc(len*sizeof(char)+21);
	snprintf(name_buffer, len+21, "%s_%s.%s", ffmpegc->basename, middle, ffmpegc->extension);
	
	if (url_fopen(&ffmpegc->oc->pb, name_buffer, URL_WRONLY) < 0) {
		EPRINT("Could not open '%s'\n", name_buffer);
		goto error0;
	}
	av_set_parameters(ffmpegc->oc, NULL);
	av_write_header(ffmpegc->oc);
	return;
 error0:

#ifdef HAVE_CONVERT
	if(ffmpegc->conv)
		convert_end(ffmpegc->conv);
#endif
	free(fct);
	free(ffmpegc);
}


struct file_core_t * ffmpegfile_core_init(char ** args, unsigned int format, ntime_t vrate) {
	struct file_core_t * fct;
	struct ffmpegfile_core_t * ffmpegc;
	int use_video = 0, use_audio = 0;
	int nb_streams = 0;
	int i;
	int len;
	char *inp = NULL;
	char *ext = NULL;
 	int arate = -1;
 	int abits = -1;
 	int achans = 1;
 	int width = -1;
 	int height = -1;
 	char *fname = NULL;
 	char filename[1024];
 	char *ff = NULL;
 	char *vc = NULL;
 	char *ac = NULL;
 	char *vhook = NULL;
	unsigned int split = NV_ROTATE_SIZE;
 	NV_ARGS;
 	
 	av_register_all();
 	
 	fct = (struct file_core_t *)malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		PERROR("Could no malloc new file core");
 		return NULL;
 	}
 	ffmpegc = (struct ffmpegfile_core_t *)malloc(sizeof(struct ffmpegfile_core_t));
 	if(!ffmpegc) {
 		PERROR("Could no malloc new ffmpeg core");
 		free(fct);
 		return NULL;
 	}
 	memset(ffmpegc, 0, sizeof(struct ffmpegfile_core_t));
 	memcpy(fct, &ffmpegfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)ffmpegc;
 	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-ff") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ff.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVOutputFormat *f;
				MPRINT("Listing FFMPEG file formats:\n");
				for(f = first_oformat; f; f = f->next) {
					MPRINT("  %s\n", f->name);
				}
				goto usage;
			}
			ff = NV_THIS_ARG;
		}else if(strcmp(NV_THIS_ARG, "-t") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -t.\n");
				goto usage;
			}
			split = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vc") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vc.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVCodec *p;
				MPRINT("Listing FFMPEG video codecs:\n");
				for(p = first_avcodec; p; p = p->next) {
					if ((p->encode) && (p->type == CODEC_TYPE_VIDEO))
						MPRINT("  %s\n", p->name);
				}
				goto usage;
			}
			vc = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-input") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -input.\n");
				goto usage;
			}
			inp = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-ext") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ext.\n");
				goto usage;
			}
			ext = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-ac") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -ac.\n");
				goto usage;
			}
			if(strcmp(NV_THIS_ARG, "help") == 0) {
				AVCodec *p;
				MPRINT("Listing FFMPEG audio codecs:\n");
				for(p = first_avcodec; p; p = p->next) {
					if ((p->encode) && (p->type == CODEC_TYPE_AUDIO))
						MPRINT("  %s\n", p->name);
				}
				goto usage;
			}
			ac = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		} else if(strcmp(NV_THIS_ARG, "-vhook") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vhook.\n");
				goto usage;
			}
			vhook = NV_THIS_ARG;
		}
	}
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}

	ffmpegc->basename = NULL;
	if(strncmp(fname, "split://", 8) == 0) {
		char middle[16];
		char *name_buffer;
		struct tm *ptr;
		time_t tm;
		
		if (!ext)
		{
			if ((inp) && (strcmp(inp, "radio") == 0))
				snprintf(ffmpegc->extension, 4, "mp3");
			else snprintf(ffmpegc->extension, 4, "avi");
		}else{

			/* FIXME: Need to check if the extension is larger than
			   3 chars */
			snprintf(ffmpegc->extension, 4, ext);
		}
		ffmpegc->split_size = split*1048576;
		IPRINT("Split Size is %d\n", ffmpegc->split_size);
		len = strlen(fname+8) + 1;
		ffmpegc->basename = (char *)malloc(len*sizeof(char));
		strcpy(ffmpegc->basename, (fname+8));
		
		tm = time(NULL);
		ptr = gmtime(&tm);
		strftime(middle, 16, "%Y%m%dT%H%M%S", ptr);
		name_buffer = (char *)malloc(len+21);
		snprintf(name_buffer, len+21, "%s_%s.%s", ffmpegc->basename, middle, ffmpegc->extension);
		fname = name_buffer;
	}
	if((width&15) || (height&15)) {
		EPRINT("Width and height must be multiples of 16.\n");
		goto usage;
	}

 	if(format != NV_YUV420) {
#ifdef HAVE_CONVERT
 		IPRINT("No native format found - initialise format conversion\n");
 		ffmpegc->conv = convert_new(format, NV_YUV420, width, height, vhook);
 		if(!ffmpegc->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto usage;
 		}
#else
		EPRINT("No native format found, and conversion not available - giving up.\n");
		goto usage;
#endif
 	}

#ifdef HAVE_CONVERT
	if((!ffmpegc->conv) && (vhook)) {
		IPRINT("Instantiating conversion function for vhooks...\n");
 		ffmpegc->conv = convert_new(format, format, width, height, vhook);
 		if(!ffmpegc->conv) {
 			EPRINT("Could not init format conversion core!\n");
 			goto usage;
 		}
	}
#endif
	if(strcmp(fname, "-") == 0) {
		snprintf(filename, 1024, "pipe:");
	}
	snprintf(filename, 1024, "%s", fname);
	IPRINT("File: %s\n", filename);

	ffmpegc->oc = av_alloc_format_context();
 	
	if(ff)
		ffmpegc->oc->oformat = guess_format(ff, NULL, NULL);
	else
		ffmpegc->oc->oformat = guess_format(NULL, filename, NULL);
	if(!ffmpegc->oc->oformat) {
		EPRINT("Could not match format for %s:%s\n", ff, fname);
		goto error0;
	}

	if((ff) && (strcmp(ff, "ffm") == 0) && (strncmp(fname, "http:", 5) == 0)) {
		int err;
		AVFormatContext *ic;
		
		IPRINT("Starting http streaming...\n");
		err = av_open_input_file(&ic, fname, NULL, FFM_PACKET_SIZE, NULL);
		if(err < 0) {
			EPRINT("Could not open input stream '%s'\n", fname);
			goto error0;
		}
		IPRINT("Requested %d streams\n", ic->nb_streams);
		if(ic->nb_streams > 2) {
			EPRINT("ffmpegrec can only stream 1 video and 1 audio stream (max 2 streams)\n");
			goto error0;
		}
		for(nb_streams = 0; nb_streams < ic->nb_streams; nb_streams++) {
			AVStream *st;
			AVCodecContext *c;
			c = &ic->streams[nb_streams]->codec;
			if(c->codec_type == CODEC_TYPE_VIDEO) {
				IPRINT("Stream %d is video\n", nb_streams);
				IPRINT(" Video frame rate %d (%d)\n", c->frame_rate, (int)(1000000000.0/(double)vrate));
 				if(c->frame_rate != (int)(1000000000.0/(double)vrate)) {
 					EPRINT("ffmpegrec and ffserver frame rate differ - try '-vr %.2f'\n", (float)c->frame_rate);
 					goto error0;
 				}
				IPRINT(" Video width %d (%d)\n", c->width, width);
 				if(c->width != width) {
 					EPRINT("ffmpegrec and ffserver video width differ - try '-w %d'\n", c->width);
 					goto error0;
 				}
				IPRINT(" Video height %d (%d)\n", c->height, height);
 				if(c->height != height) {
 					EPRINT("ffmpegrec and ffserver video height differ - try '-h %d'\n", c->height);
 					goto error0;
 				}
				if(ffmpegc->video_cc) {
					EPRINT("Video stream already configured!\n");
					goto error0;
				} else {
					st = av_new_stream(ffmpegc->oc,
							   ffmpegc->video_stream);
					
					//st = av_mallocz(sizeof(AVFormatContext));
					memcpy(st, ic->streams[nb_streams], sizeof(AVStream));
					ffmpegc->oc->streams[nb_streams] = st;
					ffmpegc->video_cc = &st->codec;
					ffmpegc->video_stream = nb_streams;
					use_video = 1;
				}
			} else if(c->codec_type == CODEC_TYPE_AUDIO) {
				IPRINT("Stream %d is audio\n", nb_streams);
				IPRINT(" Audio sample rate %d (%d)\n", c->sample_rate, arate);
 				if(c->sample_rate != arate) {
 					EPRINT("ffmpegrec and ffserver audio sample rate differ - try '-r %d'\n", c->sample_rate);
 					goto error0;
 				}
				IPRINT(" Audio channels %d (%d)\n", c->channels, achans);
 				if(c->channels != achans) {
 					EPRINT("ffmpegrec and ffserver audio channels differ - try '%s'\n", (c->channels==1)?"-m":"-s");
 					goto error0;
 				}
				if(ffmpegc->audio_cc) {
					EPRINT("Audio stream already configured!\n");
					goto error0;
				} else {
					st = av_new_stream(ffmpegc->oc,
							   ffmpegc->video_stream);
					
					//st = av_mallocz(sizeof(AVFormatContext));
					memcpy(st, ic->streams[nb_streams], sizeof(AVStream));
					ffmpegc->oc->streams[nb_streams] = st;
					ffmpegc->audio_cc = &st->codec;
					ffmpegc->audio_stream = nb_streams;
					use_audio = 1;
				}
			} else {
				EPRINT("Stream %d is unknown!\n", nb_streams);
				goto error0;
			}
		}
		av_close_input_file(ic);
	} else {
		use_video = ffmpegc->oc->oformat->video_codec != CODEC_ID_NONE;
		use_audio = ffmpegc->oc->oformat->audio_codec != CODEC_ID_NONE;
		
		IPRINT("Got format : %s (%d,%d)\n", ffmpegc->oc->oformat->name, use_video, use_audio);
		
		if(use_video){
			AVCodec *codec;
			AVStream *st;
			
			ffmpegc->video_stream = nb_streams;
			st = av_new_stream(ffmpegc->oc, ffmpegc->video_stream);
			
			if(!st){
				EPRINT("Error mallocing AVstream\n");
				goto error0;
			}
			avcodec_get_context_defaults(&st->codec);
		
			if(vc) {
				codec = avcodec_find_encoder_by_name(vc);
				ffmpegc->oc->oformat->video_codec = codec->id;
			}

			ffmpegc->video_cc = &st->codec;
			ffmpegc->video_cc->codec_id = ffmpegc->oc->oformat->video_codec;
			IPRINT("Video codec id %d\n", ffmpegc->video_cc->codec_id);
			ffmpegc->video_cc->codec_type = CODEC_TYPE_VIDEO;
			ffmpegc->video_cc->frame_rate = (int)(1000000000.0/(double)vrate);
			ffmpegc->video_cc->width = width;
			ffmpegc->video_cc->height = height;
			if(ffmpeg_setparm(args, 1, ffmpegc->video_cc) < 0) {
				EPRINT("Error setting ffmpeg video parameters.\n");
				goto usage;
			}
			ffmpegc->oc->streams[nb_streams++] = st;
		} else {
			fct->file_core_video = ffmpegfile_core_novideo;
			fct->file_core_fastvideo = ffmpegfile_core_novideo;
			fct->file_core_nullvideo = ffmpegfile_core_nonullvideo;
			DPRINT("Not using video\n");
		}

		if(use_audio){
			AVCodec *codec;
			AVStream *st;

			ffmpegc->audio_stream = nb_streams;
			st = av_new_stream(ffmpegc->oc,
					   ffmpegc->audio_stream);
			if(!st){
				EPRINT("Error mallocing AVstream\n");
				goto error0;
			}
			avcodec_get_context_defaults(&st->codec);
		
			if(ac) {
				codec = avcodec_find_encoder_by_name(ac);
				ffmpegc->oc->oformat->audio_codec = codec->id;
			}
			ffmpegc->audio_cc = &st->codec;
			ffmpegc->audio_cc->codec_id = ffmpegc->oc->oformat->audio_codec;
			IPRINT("Audio codec id %d\n", ffmpegc->audio_cc->codec_id);
			ffmpegc->audio_cc->codec_type = CODEC_TYPE_AUDIO;
			ffmpegc->audio_cc->sample_rate = arate;
			ffmpegc->audio_cc->channels = achans;
			ffmpegc->audio_cc->sample_fmt = SAMPLE_FMT_S16;
			if(ffmpeg_setparm(args, 0, ffmpegc->audio_cc) < 0) {
				EPRINT("Error setting ffmpeg audio parameters.\n");
				goto usage;
			}
			ffmpegc->oc->streams[nb_streams++] = st;
		} else {
			fct->file_core_audio = ffmpegfile_core_noaudio;
			DPRINT("Not using audio\n");
		
		}
	}
	if(!nb_streams) {
		EPRINT("No audio or video selected\n");
		goto error0;
	}

	ffmpegc->oc->nb_streams = nb_streams;

	if(use_video) {
		ffmpegc->pic = avcodec_alloc_frame();
		if(!ffmpegc->pic) {
			EPRINT("Out of memory allocating picture buffers\n");
			goto usage;
		}
		ffmpegc->pic->linesize[0] = width;
		ffmpegc->pic->linesize[1] = width / 2;
		ffmpegc->pic->linesize[2] = width / 2;
	}

	if(nvcore_verbose>NV_VERBOSE_QUIET)	
		dump_format(ffmpegc->oc, 1, filename, 1);
	
	if (url_fopen(&ffmpegc->oc->pb, filename, URL_WRONLY) < 0) {
		EPRINT("Could not open '%s'\n", filename);
		goto error0;
	}

	for(i=0; i<nb_streams; i++){
		AVCodec *codec;
		AVStream *st;

		st = ffmpegc->oc->streams[i];
		codec = avcodec_find_encoder(st->codec.codec_id);
		if(!codec){
			EPRINT("Could not find codec\n");
			goto error0;
		}
		if(avcodec_open(&st->codec, codec) < 0){
			EPRINT("Cannot open avencoder %s\n", codec->name);
			goto error0;
		}
	}

	if(use_audio){
		ffmpegc->frame_bytes = 2 * achans * ffmpegc->oc->streams[ffmpegc->audio_stream]->codec.frame_size;
		ffmpegc->audio_buffer = malloc(ffmpegc->frame_bytes);
		if(!ffmpegc->audio_buffer) {
			EPRINT("Error mallocing audio_buffer\n");
			goto error0;
		}
		ffmpegc->audio_out = malloc(ffmpegc->frame_bytes);
		if(!ffmpegc->audio_out) {
			EPRINT("Error mallocing audio_out\n");
			goto error0;
		}
	}
	av_set_parameters(ffmpegc->oc, NULL);
	av_write_header(ffmpegc->oc);
	
	return fct;

 error0:
	free(fct);
	free(ffmpegc);
	return NULL;
 usage:
	EPRINT("Error parsing options.  Please run without any command\n");
	EPRINT("line options for detailed usage information.\n");
	
	free(fct);
	free(ffmpegc);
	return NULL;
}

void ffmpegfile_core_close(struct file_core_t * fct)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int i;
	
	if(!ffmpegc)
		return;
	
	for(i = 0; i < ffmpegc->oc->nb_streams; i++)
		avcodec_close(&ffmpegc->oc->streams[i]->codec);
	
	av_write_trailer(ffmpegc->oc);
	
	if(ffmpegc->pic)
		free(ffmpegc->pic);
	if(ffmpegc->audio_out)
		free(ffmpegc->audio_out);
	if(ffmpegc->audio_buffer)
		free(ffmpegc->audio_buffer);
#ifdef HAVE_CONVERT
	if(ffmpegc->conv) {
		convert_end(ffmpegc->conv);
	}
#endif
	free(ffmpegc);
	free(fct);
}

void ffmpegfile_core_nullvideo(struct file_core_t * fct)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	
	if(!ffmpegc)
		return;
	av_write_frame(ffmpegc->oc, NULL);
}

void ffmpegfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int ret;
	AVPacket pkt;

	if(!ffmpegc)
		return;

#ifdef HAVE_CONVERT
	if(ffmpegc->conv) {
		buf = convert_convert(ffmpegc->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif
	ffmpegc->pic->data[0] = buf[0];
	ffmpegc->pic->data[1] = buf[1];
	ffmpegc->pic->data[2] = buf[2];

	if(ffmpegc->basename
	   && (fct->file_core_size(fct) > ffmpegc->split_size)
	   && (ffmpegc->video_cc->coded_frame->key_frame)) 
		/* Must wait a keyframe */
		ffmpegfile_core_rotate(fct);
	
	ret = avcodec_encode_video(ffmpegc->video_cc, 
				   ffmpegc->video_buffer, 
				   sizeof(ffmpegc->video_buffer), 
				   ffmpegc->pic);
	if(ret < 0) {
		EPRINT("Error encoding frame\n");
		av_write_frame(ffmpegc->oc, NULL);
	} else {
		av_init_packet(&pkt);
		pkt.stream_index= ffmpegc->video_stream;
		pkt.data= (uint8_t *)ffmpegc->video_buffer;
		pkt.size= ret;
		pkt.pts = ffmpegc->video_cc->coded_frame->pts;
		if(ffmpegc->video_cc->coded_frame->key_frame)
			pkt.flags |= PKT_FLAG_KEY;
		
		if(av_write_frame(ffmpegc->oc, &pkt) != 0)
			EPRINT("Error while writing video frame\n");
	}
}

void ffmpegfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct ffmpegfile_core_t * ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	int c;
	AVPacket pkt;
	
	if(!ffmpegc)
		return;
	
	if ((ffmpegc->oc->oformat->video_codec == CODEC_ID_NONE)
	    && ffmpegc->basename
	    && (fct->file_core_size(fct) > ffmpegc->split_size))
		
		ffmpegfile_core_rotate(fct);
	
	while((ffmpegc->abuf_used + n) >= ffmpegc->frame_bytes) {
		c = ffmpegc->frame_bytes - ffmpegc->abuf_used;
		memcpy(ffmpegc->audio_buffer + ffmpegc->abuf_used, buf, c);
		n -= c;
		buf += c;
		c = avcodec_encode_audio(ffmpegc->audio_cc,
					 ffmpegc->audio_out,
					 ffmpegc->frame_bytes,
					 (short *)ffmpegc->audio_buffer);
		
		av_init_packet(&pkt);
		pkt.stream_index= ffmpegc->audio_stream;
		pkt.pts = ffmpegc->audio_cc->coded_frame->pts;
		pkt.data = (uint8_t *)ffmpegc->audio_out;
		pkt.size = c;
		
		if (av_write_frame(ffmpegc->oc, &pkt) != 0)
			EPRINT("Error writing audio frame\n");
		ffmpegc->abuf_used = 0;
	}
	if(n) {
		memcpy(ffmpegc->audio_buffer + ffmpegc->abuf_used, buf, n);
		ffmpegc->abuf_used += n;
	}
}
long ffmpegfile_core_size(struct file_core_t * fct)
{
	/* Get the file size */
	struct ffmpegfile_core_t *ffmpegc = (struct ffmpegfile_core_t *)fct->priv;
	return url_filesize(url_fileno(&ffmpegc->oc->pb));
}

void ffmpegfile_core_sync(struct file_core_t * fct)
{
	return;
}

static unsigned int ffmpegfile_core_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};
//static unsigned int ffmpegfile_core_formats[] = {NV_RGB24, NV_END_FORMAT};
//static unsigned int ffmpegfile_core_formats[] = {NV_YUV422, NV_END_FORMAT};

unsigned int *ffmpegfile_core_format(char **args)
{
	return ffmpegfile_core_formats;
}

void ffmpegfile_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	return;
}

struct file_core_t ffmpegfile_core = {
	file_core_size: ffmpegfile_core_size,
	file_core_init: ffmpegfile_core_init,
	file_core_close: ffmpegfile_core_close,
	file_core_nullvideo: ffmpegfile_core_nullvideo,
	file_core_fastvideo: ffmpegfile_core_video,
	file_core_video: ffmpegfile_core_video,
	file_core_audio: ffmpegfile_core_audio,
	file_core_sync: ffmpegfile_core_sync,
	file_core_format: ffmpegfile_core_format,
	file_core_overrides: ffmpegfile_core_overrides,
	file_core_usage: ffmpegfile_core_usage,
	priv: NULL
};
